@@ -234,8 +234,9 @@ static Bool RADEONCheckTexturePOT(PicturePtr pPict, Bool canTile)
     int w = pPict->pDrawable->width;
     int h = pPict->pDrawable->height;
 
-    if (pPict->repeat && ((w & (w - 1)) != 0 || (h & (h - 1)) != 0) &&
-	!(!pPict->transform && canTile))
+    if (pPict->repeat && pPict->repeatType != RepeatPad &&
+	((w & (w - 1)) != 0 || (h & (h - 1)) != 0) &&
+	!(pPict->repeatType == RepeatNormal && !pPict->transform && canTile))
 	RADEON_FALLBACK(("NPOT repeating %s unsupported (%dx%d), transform=%d\n",
 			 canTile ? "source" : "mask", w, h, pPict->transform != 0));
 
@@ -283,7 +284,7 @@ static Bool RADEONSetupSourceTile(PicturePtr pPict,
     info->accel_state->need_src_tile_x = info->accel_state->need_src_tile_y = FALSE;
     info->accel_state->src_tile_width = info->accel_state->src_tile_height = 65536; /* "infinite" */
 	    
-    if (pPict->repeat) {
+    if (pPict->repeat && pPict->repeatType != RepeatPad) {
 	Bool badPitch = needMatchingPitch && !RADEONPitchMatches(pPix);
 	
 	int w = pPict->pDrawable->width;
@@ -296,7 +297,12 @@ static Bool RADEONSetupSourceTile(PicturePtr pPict,
 	} else {
 	    info->accel_state->need_src_tile_x = (w & (w - 1)) != 0 || badPitch;
 	    info->accel_state->need_src_tile_y = (h & (h - 1)) != 0;
-	    
+
+	    if ((info->accel_state->need_src_tile_x ||
+		 info->accel_state->need_src_tile_y) &&
+		pPict->repeatType != RepeatNormal)
+		RADEON_FALLBACK(("Can only tile RepeatNormal at this time\n"));
+
 	    if (!canTile1d)
 		info->accel_state->need_src_tile_x =
 		    info->accel_state->need_src_tile_y =
@@ -356,7 +362,7 @@ static Bool FUNC_NAME(R100TextureSetup)(PicturePtr pPict, PixmapPtr pPix,
     uint32_t txfilter, txformat, txoffset, txpitch;
     int w = pPict->pDrawable->width;
     int h = pPict->pDrawable->height;
-    Bool repeat = pPict->repeat &&
+    Bool repeat = pPict->repeat && pPict->repeatType != RepeatPad &&
 	!(unit == 0 && (info->accel_state->need_src_tile_x || info->accel_state->need_src_tile_y));
     int i;
     ACCEL_PREAMBLE();
@@ -368,7 +374,7 @@ static Bool FUNC_NAME(R100TextureSetup)(PicturePtr pPict, PixmapPtr pPix,
 	RADEON_FALLBACK(("Bad texture offset 0x%x\n", (int)txoffset));
     if ((txpitch & 0x1f) != 0)
 	RADEON_FALLBACK(("Bad texture pitch 0x%x\n", (int)txpitch));
-    
+
     for (i = 0; i < sizeof(R100TexFormats) / sizeof(R100TexFormats[0]); i++)
     {
 	if (R100TexFormats[i].fmt == pPict->format)
@@ -403,8 +409,22 @@ static Bool FUNC_NAME(R100TextureSetup)(PicturePtr pPict, PixmapPtr pPix,
 	RADEON_FALLBACK(("Bad filter 0x%x\n", pPict->filter));
     }
 
-    if (repeat)
-      txfilter |= RADEON_CLAMP_S_WRAP | RADEON_CLAMP_T_WRAP;
+    if (repeat) {
+	switch (pPict->repeatType) {
+	case RepeatNormal:
+	    txfilter |= RADEON_CLAMP_S_WRAP | RADEON_CLAMP_T_WRAP;
+	    break;
+	case RepeatPad:
+	    txfilter |= RADEON_CLAMP_S_CLAMP_LAST | RADEON_CLAMP_T_CLAMP_LAST;
+	    break;
+	case RepeatReflect:
+	    txfilter |= RADEON_CLAMP_S_MIRROR | RADEON_CLAMP_T_MIRROR;
+	    break;
+	case RepeatNone:
+	    /* Nothing to do */
+	    break;
+	}
+    }
 
     BEGIN_ACCEL(5);
     if (unit == 0) {
@@ -539,7 +559,7 @@ static Bool FUNC_NAME(R100PrepareComposite)(int op,
 	return FALSE;
 
     if (pDstPicture->format == PICT_a8 && RadeonBlendOp[op].dst_alpha)
-	RADEON_FALLBACK("Can't dst alpha blend A8\n");
+	RADEON_FALLBACK(("Can't dst alpha blend A8\n"));
 
     if (pMask)
 	info->accel_state->has_mask = TRUE;
@@ -678,7 +698,7 @@ static Bool FUNC_NAME(R200TextureSetup)(PicturePtr pPict, PixmapPtr pPix,
     uint32_t txfilter, txformat, txoffset, txpitch;
     int w = pPict->pDrawable->width;
     int h = pPict->pDrawable->height;
-    Bool repeat = pPict->repeat &&
+    Bool repeat = pPict->repeat && pPict->repeatType != RepeatPad &&
 	!(unit == 0 && (info->accel_state->need_src_tile_x || info->accel_state->need_src_tile_y));
     int i;
     ACCEL_PREAMBLE();
@@ -727,8 +747,22 @@ static Bool FUNC_NAME(R200TextureSetup)(PicturePtr pPict, PixmapPtr pPix,
 	RADEON_FALLBACK(("Bad filter 0x%x\n", pPict->filter));
     }
 
-    if (repeat)
-      txfilter |= R200_CLAMP_S_WRAP | R200_CLAMP_T_WRAP;
+    if (repeat) {
+	switch (pPict->repeatType) {
+	case RepeatNormal:
+	    txfilter |= R200_CLAMP_S_WRAP | R200_CLAMP_T_WRAP;
+	    break;
+	case RepeatPad:
+	    txfilter |= R200_CLAMP_S_CLAMP_LAST | R200_CLAMP_T_CLAMP_LAST;
+	    break;
+	case RepeatReflect:
+	    txfilter |= R200_CLAMP_S_MIRROR | R200_CLAMP_T_MIRROR;
+	    break;
+	case RepeatNone:
+	    /* Nothing to do */
+	    break;
+	}
+    }
 
     BEGIN_ACCEL(6);
     if (unit == 0) {
@@ -847,7 +881,7 @@ static Bool FUNC_NAME(R200PrepareComposite)(int op, PicturePtr pSrcPicture,
 	return FALSE;
 
     if (pDstPicture->format == PICT_a8 && RadeonBlendOp[op].dst_alpha)
-	RADEON_FALLBACK("Can't dst alpha blend A8\n");
+	RADEON_FALLBACK(("Can't dst alpha blend A8\n"));
 
     if (pMask)
 	info->accel_state->has_mask = TRUE;
@@ -1061,17 +1095,38 @@ static Bool FUNC_NAME(R300TextureSetup)(PicturePtr pPict, PixmapPtr pPix,
     info->accel_state->texW[unit] = w;
     info->accel_state->texH[unit] = h;
 
-    if (pPict->repeat && !(unit == 0 && info->accel_state->need_src_tile_x))
-      txfilter = R300_TX_CLAMP_S(R300_TX_CLAMP_WRAP);
-    else
-      txfilter = R300_TX_CLAMP_S(R300_TX_CLAMP_CLAMP_GL);
+    txfilter = (unit << R300_TX_ID_SHIFT);
 
-    if (pPict->repeat && !(unit == 0 && info->accel_state->need_src_tile_y))
-      txfilter |= R300_TX_CLAMP_T(R300_TX_CLAMP_WRAP);
-    else
-      txfilter |= R300_TX_CLAMP_T(R300_TX_CLAMP_CLAMP_GL);
+    if (pPict->repeat) {
+	switch (pPict->repeatType) {
+	case RepeatNormal:
+	    if (unit != 0 || !info->accel_state->need_src_tile_x)
+		txfilter |= R300_TX_CLAMP_S(R300_TX_CLAMP_WRAP);
+	    else
+		txfilter |= R300_TX_CLAMP_S(R300_TX_CLAMP_CLAMP_GL);
 
-    txfilter |= (unit << R300_TX_ID_SHIFT);
+	    if (unit != 0 || !info->accel_state->need_src_tile_y)
+		txfilter |= R300_TX_CLAMP_T(R300_TX_CLAMP_WRAP);
+	    else
+		txfilter |= R300_TX_CLAMP_T(R300_TX_CLAMP_CLAMP_GL);
+
+	    break;
+	case RepeatPad:
+	    txfilter |= R300_TX_CLAMP_S(R300_TX_CLAMP_CLAMP_LAST) |
+		        R300_TX_CLAMP_T(R300_TX_CLAMP_CLAMP_LAST);
+	    break;
+	case RepeatReflect:
+	    txfilter |= R300_TX_CLAMP_S(R300_TX_CLAMP_MIRROR) |
+		        R300_TX_CLAMP_T(R300_TX_CLAMP_MIRROR);
+	    break;
+	case RepeatNone:
+	    txfilter |= R300_TX_CLAMP_S(R300_TX_CLAMP_CLAMP_GL) |
+		        R300_TX_CLAMP_T(R300_TX_CLAMP_CLAMP_GL);
+	    break;
+	}
+    } else
+	txfilter |= R300_TX_CLAMP_S(R300_TX_CLAMP_CLAMP_GL) |
+	            R300_TX_CLAMP_T(R300_TX_CLAMP_CLAMP_GL);
 
     switch (pPict->filter) {
     case PictFilterNearest:
@@ -1969,7 +2024,7 @@ static void FUNC_NAME(RadeonCompositeTile)(PixmapPtr pDst,
     RINFO_FROM_SCREEN(pDst->drawable.pScreen);
     int vtx_count;
     xPointFixed srcTopLeft, srcTopRight, srcBottomLeft, srcBottomRight;
-    xPointFixed maskTopLeft, maskTopRight, maskBottomLeft, maskBottomRight;
+    static xPointFixed maskTopLeft, maskTopRight, maskBottomLeft, maskBottomRight;
     ACCEL_PREAMBLE();
 
     ENTER_DRAW(0);
@@ -1986,31 +2041,32 @@ static void FUNC_NAME(RadeonCompositeTile)(PixmapPtr pDst,
     srcBottomRight.x = IntToxFixed(srcX + w);
     srcBottomRight.y = IntToxFixed(srcY + h);
 
-    maskTopLeft.x     = IntToxFixed(maskX);
-    maskTopLeft.y     = IntToxFixed(maskY);
-    maskTopRight.x    = IntToxFixed(maskX + w);
-    maskTopRight.y    = IntToxFixed(maskY);
-    maskBottomLeft.x  = IntToxFixed(maskX);
-    maskBottomLeft.y  = IntToxFixed(maskY + h);
-    maskBottomRight.x = IntToxFixed(maskX + w);
-    maskBottomRight.y = IntToxFixed(maskY + h);
-
     if (info->accel_state->is_transform[0]) {
 	transformPoint(info->accel_state->transform[0], &srcTopLeft);
 	transformPoint(info->accel_state->transform[0], &srcTopRight);
 	transformPoint(info->accel_state->transform[0], &srcBottomLeft);
 	transformPoint(info->accel_state->transform[0], &srcBottomRight);
     }
-    if (info->accel_state->is_transform[1]) {
-	transformPoint(info->accel_state->transform[1], &maskTopLeft);
-	transformPoint(info->accel_state->transform[1], &maskTopRight);
-	transformPoint(info->accel_state->transform[1], &maskBottomLeft);
-	transformPoint(info->accel_state->transform[1], &maskBottomRight);
-    }
 
-    if (info->accel_state->has_mask)
+    if (info->accel_state->has_mask) {
+	maskTopLeft.x     = IntToxFixed(maskX);
+	maskTopLeft.y     = IntToxFixed(maskY);
+	maskTopRight.x    = IntToxFixed(maskX + w);
+	maskTopRight.y    = IntToxFixed(maskY);
+	maskBottomLeft.x  = IntToxFixed(maskX);
+	maskBottomLeft.y  = IntToxFixed(maskY + h);
+	maskBottomRight.x = IntToxFixed(maskX + w);
+	maskBottomRight.y = IntToxFixed(maskY + h);
+
+	if (info->accel_state->is_transform[1]) {
+	    transformPoint(info->accel_state->transform[1], &maskTopLeft);
+	    transformPoint(info->accel_state->transform[1], &maskTopRight);
+	    transformPoint(info->accel_state->transform[1], &maskBottomLeft);
+	    transformPoint(info->accel_state->transform[1], &maskBottomRight);
+	}
+
 	vtx_count = 6;
-    else
+    } else
 	vtx_count = 4;
 
     FUNC_NAME(RADEONWaitForVLine)(pScrn, pDst, RADEONBiggerCrtcArea(pDst), dstY, dstY + h, info->accel_state->vsync);
